// Copyright 2019 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// Code generated by generate-protos. DO NOT EDIT.

package fieldnum

// Field numbers for google.protobuf.Any.
const (
	Any_TypeUrl = 1 // optional string
	Any_Value   = 2 // optional bytes
)
